"""argetype pipelines module

I have put a lot of philosophical thought into what is the difference
between a pipeline and a task. In the implementation of this package
they both implement the RunInterface: they are supposed to run something
and reach a state of completion at some point.

A Task (TaskBase), when it explicitly defines all its dependencies, can 
be considered a pipeline indirectly, and it is not necessary to wrap it 
in a Pipeline class (PbBase).

A Pipeline class is useful when some glue code is necesarry between
different tasks:
    - output and input of subsequent Tasks must be mapped
    - a Task must be applied to a range of samples

Todo:
    - extracting task settings

Example:
    >>> from argetype.pipelines import PbBase, step
    ... from argetype.tasks import TaskBase
    ... 
    ... class TaskA(TaskBase):
    ...     a: int = 1
    ... 
    ...     def out1(self) -> int:
    ...         return self.a**2
    ... 
    ... class TaskB(TaskBase):
    ...     dependence: TaskBase
    ...     b: int = 2
    ... 
    ...     def out2(self) -> int:
    ...         return self.dependence.a * self.b
    ... 
    ... class TaskC(TaskBase):
    ...     taskB: TaskB
    ... 
    ... class Pipeline(PbBase):
    ...     taskA: TaskA = STEP()
    ...     taskB: TaskB = STEP(input={'dependence': 'taskA'})
    ...     end:   TaskC = STEP() # will automatically serve taskB
    ... 
    ... pipeline = Pipeline()
    ... pipeline.run()

"""
from .tasks import RunInterface
from itertools import count

class StepBase(object):
    lastNumber = count(0)
    
    def __init__(self, number = None, input = {}, output = {}):
        self.number = number if number else next(self.lastNumber)
        self.input = input
        self.output = output

    def __call__(self, pipeline, taskName):
        # TODO check previous ran tasks, and see if they need to be added
        task = pipeline._tasks[taskName](parse=False)
        if not task.completed():
            task(**{k:pipeline._tasks[v] for k,v in self.input.items()})
            task.run()
            pipeline._tasks[taskName] = task

STEP = StepBase

class PbBase(RunInterface):
    def __init__(self):
        import typing
        import inspect
        self._tasks = typing.get_type_hints(self)
        self.steps = list(self._tasks)
        self._step_management = {
            k:v for k,v in vars(type(self)).items()
            if k in self._tasks
        }
        for step in self.steps:
            if step not in self._step_management:
                self._step_management[step] = STEP(number=-1)
        self._ran = False

    def run(self):
        self._run_started = True
        for step in self.steps:
            self._step_management[step](
                pipeline = self,
                taskName = step
            )
        self._ran = True

    def completed(self):
        return self._ran



"""Types for use in argetype configs and tasks

This module establishes ways to convert to CLI
or to do type checking for the generated calls.
"""
import abc
import typing

class Archetype(abc.ABC):
    @abc.abstractmethod
    def convert_to_CL():
        pass

    @abc.abstractmethod
    def convert_to_call():
        pass

class Cacheable(abd.ABC):
    """Abstract cache for use in function return types
    that are cacheable.
    """
    @abc.abstractmethod
    def save_to_cache():
        pass

    @abc.abstractmethod
    def load_from_cache():
        pass

    @abc.abstractmethod
    def exists():
        pass

class CACHE(Cacheable):
    def __init__(self, type, location=None, protocol=None):
        self.type = type
        self.location = location
        self.protocol = protocol

    def save_to_cache(self, key, value):
        pass

    def load_from_cache(self):
        pass

    def exists(self, key):
        # TODO Saving not yet implemented
        return False



"""argetype tasks module takes the ConfigBase
to build a inheritable TaskBase class around it.

Todo:
    - dependency settings inheritance and setting

Example:
    >>> from argtype.tasks import TaskBase
    ... class TaskDep(TaskBase):
    ...     a: str = '/tmp/file1'
    ... 
    ...     def generate_output(self) -> str:
    ...         return self.a    
    ... 
    ... class Task(TaskBase):    
    ...     # Task settings
    ...     a: int = 10
    ...     b: str = 'a'
    ... 
    ...     def generate_output1(self, task_dependence1: TaskDep) -> int:
    ...         print(task_dependence1.a)
    ...         return 0
    ...     
    ...     def generate_output2(self) -> str:
    ...         with self.env('sh') as env:
    ...             env.exec('which python')
    ...             return env.output
    ... 
    ...     def generate_output3(self) -> str:
    ...         with self.env('py') as env:
    ...             env.exec(f'''
    ...             for i in range({self.a}):
    ...                 print(i)
    ...             ''')
    ...             return env.output
    ... 
    ... task = Task()
    ... task.run()
    ... print(task._input, task._output)

""" 

import os
import abc
import typing
import inspect
from collections import OrderedDict
from argetype import ConfigBase

class RunInterface(abc.ABC):
    @abc.abstractmethod
    def run(self):
        pass

    @abc.abstractmethod
    def completed(self):
        pass

class TaskBase(ConfigBase, RunInterface):
    def __init__(self, parse=False):
        # Parsing at object creation is not that useful
        # Only when a task runs should it have all its settings.
        super().__init__(parse=parse)
        self.taskprep()
        self._input = {}
        self._output = {}
        
    def taskprep(self):
        cls = type(self)

        # Small utility function to sort members in order of appearance
        def memberline(m):
            try:
                return m[1].__code__.co_firstlineno
            except AttributeError:
                try:
                    # Interactive code may need extra attribute
                    return m[1].__func__.__code__.co_firstlineno
                except:
                    return -1
    
        self._output_functions = OrderedDict([
            (name, inspect.signature(fun))
            for name, fun in sorted(
                    inspect.getmembers(cls, predicate=inspect.isfunction),
                    key=memberline
            )
            if typing.get_type_hints(fun)
        ])

    def run(self, fail=True):
        # Task can declare a verbose attribute used in this run
        try: verbose = self.__getattribute__('verbose')
        except AttributeError: verbose = False

        # Run task subtask methods
        for fn in self._output_functions:
            if fn in self._output:
                if verbose: print(fn, 'already generated output')
                continue
            return_type = self._output_functions[fn].return_annotation
            # TODO could check here if return type is correct in earlier generated output
            parameters = self._output_functions[fn].parameters
            function_inputs = {}
            for dependency in parameters:
                if dependency in self._input:
                    function_inputs[dependency] = self._input[dependency]
                # If annotation present for dependency it should be a task class
                elif parameters[dependency].annotation is not inspect._empty:
                    # Instantiate dependency and run
                    deptask = parameters[dependency].annotation()
                    deptask.run()
                    self._input[dependency] = deptask
                    function_inputs[dependency] = deptask
                # If no annotation it could be the output generated from one of the task methods
                elif dependency in self._output:
                    function_inputs[dependency] = self._output[dependency]
                # Finally, it could also be an attribute of the task class
                else:
                    try:
                        function_inputs[dependency] = self.__getattribute__(dependency)
                    except AttributeError:
                        # check if attribute simply refers to 'self' or similar
                        if dependency not in ('_', 'self', 'task'):
                            print(dependency, 'not found') #TODO make warning
            if verbose: print('Executing', fn, '...')
            return_value = self.__getattribute__(fn)(
                **function_inputs
            )
            if verbose: print('done')
            assert isinstance(return_value, return_type)
            self._output[fn] = return_value

    def env(self, environment):
        return ExecEnvironment(environment)

    def completed(self):
        return bool(self._output)

class ExecEnvironment(object):
    predefined_envs = {
        'sh': ['bash', []],
        'py': ['python', []]
    }
    
    def __init__(self, command, options=[]):
        import threading
        import shutil
        self.lock = threading.Lock()
        self.command = shutil.which(
            self.predefined_envs[command][0]
            if command in self.predefined_envs else command
        )
        self.options = (options if options else
            self.predefined_envs[command][1]
            if command in self.predefined_envs else []
        )

    def __enter__(self):
        import tempfile
        self.lock.acquire()
        self.tmpfile = tempfile.NamedTemporaryFile(mode = 'w+t', delete = False)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # TODO log issues
        os.remove(self.tmpfile.name)
        self.lock.release()

    def exec(self, script, check = True, reident=True):
        import subprocess
        import re
        
        # reident
        if reident:
            identspace = re.compile(r'.*\n(\W*)\w')
            script = script.replace('\n'+identspace.match(script).groups()[0],'\n')
            
        # write tmp script file
        try:
            self.tmpfile.write(script)
        finally:
            self.tmpfile.close()
            
        # execute
        proc = subprocess.run(
            [self.command]+self.options+[self.tmpfile.name],
            text = True, capture_output = True, check = check
        )
        self.output = proc.stdout
        self.error = proc.stderr



from setuptools import setup, find_packages

setup(name='apetype',
      version='0.0.1',
      description='Embracing builtin python modules argparse and typing',
      url='https://github.com/dicaso/apetype',
      author='Christophe Van Neste',
      author_email='christophe.vanneste@ugent.be',
      license='MIT',
      packages=find_packages(),
      python_requires='>=3.6',
      install_requires=[],
      extras_require={
          'documentation': ['Sphinx']
      },
      package_data={},
      include_package_data=True,
      zip_safe=False,
      entry_points={},
      test_suite='nose.collector',
      tests_require=['nose']
      )

# To install with symlink, so that changes are immediately available:
# pip install -e .



"""argetype module

Provides base class `Settings` that can be used
for making a derivative class that contains typed
settings, for a module or package.
"""

import typing
from collections import OrderedDict

class ConfigBase(object):
    """Takes in a list of settings, which will be exposed
    as CLI arguments. Each settings tuple should have the
    following format:
    ('--name', keyword dict for the parser.add_argument function)

    The recommended way to build a SettingsBase object, is to
    inherit from it and define the `setup` method
    (see SettingsBase.setup docstring)

    Args:
      parse (bool | list): if True, already parse arguments.
        Can also be a list that will then be passed on as args.
    """

    def __init__(self, parse=True):
        self.groups = False
        self._help = self.get_arg_docs()
        self.setup()
        self.make_call()
        self.make_parser()
        if parse:
            self.parse_args(None if isinstance(parse, bool) else parse)

    def __getitem__(self, key):
        return self.settings.__getattribute__(key)

    def __getattr__(self, name):
        if 'settings' in self.__dict__ and name in self.settings:
            return self.settings.__getattribute__(name)
        else:
            return self.__dict__[name]

    def __repr__(self):
        if 'settings' in self.__dict__:
            return str(self.settings)
        else:
            return f'<Uninitialised {self.__class__.__name__}>'

    def __str__(self):
        return self.__repr__().replace(
            'Namespace(', 'Settings:\n\n').replace(', ', '\n')[:-1]

    def setup(self):
        """Can be overwritten by inheriting classes.
        Allows defining parameters with type hints.
        Overwritten setup methods need to call `super().setup()`
        at the end.

        Example:
        >>> class Settings(SettingsBase):
        ...     def setup(_, a: int = 5, b: float = .1, c: str = 'a'):
        ...          super().setup()
        ... settings = Settings()
        """
        import inspect
        cls = type(self)
        cls_vars = vars(cls)
        # getting hints from self does not always work
        self._annotations = annotations = typing.get_type_hints(cls)
        self._annotation_groups = annotation_groups = [
            attr for attr in cls_vars if isinstance(cls_vars[attr], type)
        ]
        if annotations or annotation_groups:
            # get_attr does not work so requesting vars
            self._settings = [
                self.add_arg_format(p, cls_vars[p] if p in cls_vars else None,
                                    annotations[p], p not in cls_vars,
                                    self._help[p] if p in self._help else None
                )
                for p in annotations
            ]
            if annotation_groups:
                self.groups = True
                self._settings = OrderedDict(('default', self._settings)) if self._settings else OrderedDict()
                for annot_grp in annotation_groups:
                    grp_annotations = typing.get_type_hints(cls_vars[annot_grp])
                    grp_cls_vars = vars(cls_vars[annot_grp])
                    self._settings[annot_grp] = [
                        self.add_arg_format(
                            p, grp_cls_vars[p] if p in grp_cls_vars else None,
                            grp_annotations[p], p not in grp_cls_vars,
                            self._help[p] if p in self._help else None
                        )
                        for p in grp_annotations
                    ]
        else:
            sig = inspect.signature(self.setup)
            self._settings = [
                self.add_arg_format(
                    p, sig.parameters[p].default,
                    sig.parameters[p].annotation,
                    not(bool(sig.parameters[p].default)),
                    self._help[p] if p in self._help else None
                )
                for p in sig.parameters
            ]

    @staticmethod
    def add_arg_format(name, default, typ, positional, help):
        if typ is bool:
            return (f'--{name}', {
                    'default': default,
                    'action': 'store_const',
                    'const': not(default),
                    'help': help
            }
            )
        else:
            return (name if positional else f'--{name}', {
                    'default': default,
                    'type': typ,
                    'help': help
            }
            )

    @classmethod
    def get_arg_docs(cls):
        import re
        import inspect
        commentdoc = re.compile(r'\s*(\w+)\s*:\s*\w+.*?#(.+)')
        multilinedoc = re.compile(
            r'\s*(?P<identifier>\w+)\s*:\s*\w+.*?\n\s*(?P<delimiter>"""|\'\'\')'
            r'(?P<annot>(.*\n)+?)\s*?(?P=delimiter)',
            re.MULTILINE)
        try:
            argdocs = dict([
                commentdoc.match(line).groups()
                for line in inspect.getsourcelines(cls)[0] if commentdoc.match(line)
            ])
            # Clean if not identifier and strip
            argdocs = {arg:argdocs[arg].strip() for arg in argdocs if arg.isidentifier()}
            argdocs.update({
                match.group('identifier'):match.group('annot').replace('\n', ' ').strip()
                for match in multilinedoc.finditer(inspect.getsource(cls))
                if match.group('identifier').isidentifier()
            })
            return argdocs
        except TypeError:
            # TypeError will be raised when class is defined interactively
            # Returning an empty dict in that case
            return {}

    def set_settings(self, vardict):
        if 'self' in vardict: vardict.pop('self')
        for attr in vardict:
            # TODO check type
            self.__setattr__(attr, vardict[attr])
            # print(attr,vardict[attr])
        self.settings = vardict

    def make_call(self):
        from types import FunctionType
        variables = ', '.join([
            s[0].replace('-','') + ('' if s[1]['default'] is None else
                                     '='+repr(s[1]['default'])
            )
            for s in sorted(
                self._settings,
                key=lambda x: x[1]['default'] is None,
                reverse=True
            )
        ])
        call_code = compile(f'''def set_variables(self, {variables}):
            self.set_settings(locals())
        ''', "<string>", "exec")
        # Find code sub object:
        #for i in range(len(call_code.co_consts)):
        #    if type(call_code) is type(call_code.co_consts[i]):
        #        code = call_code.co_consts[i]
        #        break
        code_pos = 0 if call_code.co_consts[-1] is None else -4
        call_func = FunctionType(
            call_code.co_consts[code_pos], globals(), "set_variables",
            argdefs = call_code.co_consts[-1] if '=' in variables else None
        )
        setattr(self.__class__, '__call__', call_func)
        # could also orverwrite __init__, if call to super().__init__ is included, but this should be optional as otherwise it will not work for the argparse workflow side
        
    def make_parser(self, **kwargs):
        import argparse
        self.parser = argparse.ArgumentParser(**kwargs)
        if self.groups: self.group_parsers = {}
        for grp in self._settings:
            if self.groups:
                parser = self.parser.add_argument_group(grp)
                self.group_parsers[grp] = parser
            else:
                parser = self.parser
            for setting in (self._settings[grp] if self.groups else self._settings):
                parser.add_argument(
                    setting[0],
                    **setting[1]
                )
            if not self.groups:
                break  # if no groups need to break

    def parse_args(self, args):
        self.settings = self.parser.parse_args(args)
        # Overwrite default typing values
        for a in self._annotations:
            if a in self.settings:
                self.__setattr__(a, self.settings.__getattribute__(a))
        return self.settings
